package distanceFunction;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.LineNumberReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Observable;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Vector;

import org.apache.commons.lang3.mutable.MutableBoolean;
import org.apache.commons.lang3.mutable.MutableInt;

import dataStructures.Store;
import distanceFunction.random.RandomDistanceFunctions;

/**
 * Representation of a distance function.
 * 
 * @author Sepp Hartung
 * @version 1.0
 */
public class SymmetricDiscreteDistanceFunction extends Observable implements ReadOnlyViewOnSymmetricDiscreteDistanceFunction
{
	public class DistanceFunctionChange
	{
		public int i, j, b, a;

		public DistanceFunctionChange(final int i, final int j, final int b, final int a)
		{
			this.i = i;
			this.j = j;
			this.b = b;
			this.a = a;
		}

		public boolean isNegative()
		{
			return this.a < this.b;
		}

		public boolean isNeutral()
		{
			return this.a == this.b;
		}

		public boolean isPositive()
		{
			return this.a > this.b;
		}

		@Override
		public String toString()
		{
			return "DistanceFunctionChange{" + "i=" + this.i + ", j=" + this.j + ", b=" + this.b + ", a=" + this.a + '}';
		}

	}

	static class Node
	{
		public final int endInternal;
		private final int extIndex;
		private final String label;
		public final int startIndex;

		public Node(final int extIndex, final String label, final int startInternal, final int endInternal)
		{
			this.extIndex = extIndex;
			this.label = label;
			this.startIndex = startInternal;
			this.endInternal = endInternal;
		}

		public int getNumberOfNodes()
		{
			return (this.endInternal - this.startIndex) + 1;
		}

		public int getRange()
		{
			return (this.endInternal - this.startIndex) + 1;
		}

		@Override
		public String toString()
		{
			return "Node [extIndex=" + this.extIndex + ", label=" + this.label + ", startIndex=" + this.startIndex + ", endInternal="
					+ this.endInternal + "]";
		}
	}

	public final static String sep = ",";

	public static MutableInt verbosityLevelWhenReadingCM = new MutableInt(0);

	public static MutableBoolean writerUncompressedCM_File = new MutableBoolean(false);

	public static SymmetricDiscreteDistanceFunction add(final SymmetricDiscreteDistanceFunction s1, final SymmetricDiscreteDistanceFunction s2)
	{
		final int[][] d1 = s1.distanceFunction;
		final int[][] d2 = s2.distanceFunction;
		int i = 0;

		final int[][] d3 = new int[d1.length][];

		for (final int[] row : d1)
		{
			int j = 0;
			d3[i] = new int[d1[i].length];
			for (final int element : row)
			{
				d3[i][j] = d1[i][j] + d2[i][j];
				j++;
			}
			i++;
		}

		final SymmetricDiscreteDistanceFunction sum = new SymmetricDiscreteDistanceFunction(d3);
		return sum;
	}

	private static void addAllIndices(final Vector<Integer> vec, final Node node)
	{
		for (int i = node.startIndex; i <= node.endInternal; i++)
		{
			vec.add(i);
		}
	}

	public static int calcHash(final SymmetricDiscreteDistanceFunction sym)
	{
		final int PRIME = 31;

		int hash = 0;
		for (int i = 0; i < sym.getNoOfElements(); i++)
		{
			for (int j = i + 1; j < sym.getNoOfElements(); j++)
			{

				hash += PRIME * sym.get(i, j) * (i + PRIME) * (j + PRIME);
			}
		}
		return hash;
	}

	public static int[][] copyArray(final int[][] arr)
	{
		final int[][] distance = new int[arr.length][];
		for (int i = 0; i < arr.length; i++)
		{
			distance[i] = arr[i].clone();
		}
		return distance;
	}

	/**
	 * Reads cost matrix files used by Böcker et al in their Algorithmica paper.
	 * 
	 * @param filePath
	 * @return
	 * @throws IOException
	 */
	public static SymmetricDistanceFunctionIntFunction<Double> createInstanceFromCostMatrixFile(final String filePath) throws IOException
	{
		final LineNumberReader reader = new LineNumberReader(new FileReader(new File(filePath)));
		String line;

		int n = -1;
		SymmetricDistanceFunctionIntFunction<Double> sym = null;
		while ((line = reader.readLine()) != null)
		{
			if (reader.getLineNumber() == 1)
			{
				n = Integer.parseInt(line);
				sym = new SymmetricDistanceFunctionIntFunction<Double>(n);
			} else if ((reader.getLineNumber() > (1 + n)) && (reader.getLineNumber() <= (2 * n)))
			{
				final String[] arr = line.split("\t");
				for (int i = 0; i < arr.length; i++)
				{
					final int n1 = n - 1 - (reader.getLineNumber() - n - 2);
					final int n2 = n1 - (i + 1);
					sym.set(n1, n2, Double.parseDouble(arr[i]));
				}
			}
		}
		reader.close();
		return sym;
	}

	public static SymmetricDiscreteDistanceFunction createInstanceFromFile(final String filePath) throws IOException
	{
		final LineNumberReader reader = new LineNumberReader(new FileReader(new File(filePath)));
		String line;
		final Vector<int[]> results = new Vector<int[]>();
		while ((line = reader.readLine()) != null)
		{
			// ignore first (description) line
			if ((reader.getLineNumber() > 1) && (line.length() != 0))
			{
				final String[] arrS = line.split(SymmetricDiscreteDistanceFunction.sep);
				final int[] arr = new int[arrS.length];
				for (int i = 0; i < arrS.length; i++)
				{
					arr[i] = Integer.parseInt(arrS[i]);
				}
				results.add(arr);
			}
		}
		reader.close();
		return new SymmetricDiscreteDistanceFunction(results.toArray(new int[0][0]));
	}

	public static SymmetricDiscreteDistanceFunction createInstanceFromGraphFile(final String filePath) throws IOException
	{
		final LineNumberReader reader = new LineNumberReader(new FileReader(new File(filePath)));
		String line;
		final SortedSet<Integer> vertices = new TreeSet<Integer>();
		final ArrayList<Store<Integer, Integer>> edges = new ArrayList<Store<Integer, Integer>>();

		while ((line = reader.readLine()) != null)
		{
			final String[] verticesT = line.split("\\s");
			if (verticesT.length == 2)
			{
				final int v1 = Integer.parseInt(verticesT[0]);
				final int v2 = Integer.parseInt(verticesT[1]);
				edges.add(new Store<Integer, Integer>(v1, v2));
				vertices.add(v1);
				vertices.add(v2);
				System.out.println("add:" + v1 + " " + v2);
			} else
			{
				System.err.println("cannot correctly parse edge:" + line);
			}
		}
		final SymmetricDiscreteDistanceFunction distFunction = new SymmetricDiscreteDistanceFunction(vertices.size());
		final Integer[] vertexArr = vertices.toArray(new Integer[] {});

		for (int i = 0; i < distFunction.getNoOfElements(); i++)
		{
			for (int j = i + 1; j < distFunction.getNoOfElements(); j++)
			{
				distFunction.set(i, j, 2);
			}
		}
		for (final Store<Integer, Integer> edge : edges)
		{
			distFunction.set(Arrays.binarySearch(vertexArr, edge.getS1()), Arrays.binarySearch(vertexArr, edge.getS2()), 1);
		}
		return distFunction;
	}

	public static Store<Store<SymmetricDiscreteDistanceFunction, Integer>, Store<ArrayList<int[]>, ArrayList<int[]>>> createSymmetricDiscreteDistanceFunctionInstanceFromCostMatrixFile(
			final String filePath) throws IOException
	{
		if (!filePath.endsWith(".cm"))
			throw new IllegalArgumentException("expected to get *.cm file");

		final ArrayList<int[]> increaseBlocks = new ArrayList<int[]>();
		final ArrayList<int[]> decreaseBlocks = new ArrayList<int[]>();

		final LineNumberReader reader = new LineNumberReader(new FileReader(new File(filePath)));
		String line;

		int n = Integer.MAX_VALUE / 2;
		SymmetricDiscreteDistanceFunction sym = null;
		Node[] nodes = null;

		int totalReverseDRCorrect = 0;

		int internalIndex = 0;
		while (((line = reader.readLine()) != null) && (reader.getLineNumber() <= (2 * n)))
		{
			if (reader.getLineNumber() == 1)
			{
				n = Integer.parseInt(line);
				nodes = new Node[n];
			} else if (reader.getLineNumber() <= (1 + n))
			{
				final String[] arr = line.split("\\|");
				final int length = arr.length;
				final Node node = new Node((n - reader.getLineNumber()) + 1, line, internalIndex, (internalIndex + length) - 1);
				nodes[(n - reader.getLineNumber()) + 1] = node;
				internalIndex += length;
				// System.out.println(node);
			} else
			// (reader.getLineNumber() > 1 + n)
			{// lines of the input matrix
				if (sym == null)
				{
					sym = new SymmetricDiscreteDistanceFunction(internalIndex);
					for (final Node node : nodes)// add edges between merged
													// vertices
					{
						for (int i1 = node.startIndex; i1 <= node.endInternal; i1++)
						{
							for (int i2 = i1 + 1; i2 <= node.endInternal; i2++)
							{
								sym.set(i1, i2, 1);
								increaseBlocks.add(new int[] { i1, i2 });
							}
						}
					}
				}
				final String[] arr = line.split("\t");
				for (int i = 0; i < arr.length; i++)
				{
					final int externalIndex1 = n - 1 - (reader.getLineNumber() - n - 2);
					final int externalIndex2 = externalIndex1 - (i + 1);
					final double doubleValue = Double.parseDouble(arr[i]);
					final String replaceAll = arr[i];// .replaceAll("\\.0*", "");
					final long longValue = Double.valueOf(replaceAll).longValue();

					int intValue;

					if ((longValue > (Integer.MAX_VALUE / 2)) || (longValue < (Integer.MIN_VALUE / 2)) || (doubleValue > (Integer.MAX_VALUE / 2))
							|| (doubleValue < (Integer.MIN_VALUE / 2)))
					{
						intValue = doubleValue > 0 ? Integer.MAX_VALUE : Integer.MIN_VALUE;
					} else
					{

						// System.out.println("parse from:" + replaceAll + " " + Long.parseLong(replaceAll));

						intValue = Integer.parseInt(replaceAll);
					}
					final int initialIntValue = intValue;

					final Node node1 = nodes[externalIndex1];
					final Node node2 = nodes[externalIndex2];

					final int n1Range = (node1.endInternal - node1.startIndex) + 1;
					final int n2Range = (node2.endInternal - node2.startIndex) + 1;

					if ((intValue != Integer.MAX_VALUE) && (intValue != Integer.MIN_VALUE))
					{
						final int diff = (n1Range * n2Range) - Math.abs(intValue);
						if (diff < 0)
						{
							reader.close();
							throw new IllegalAccessError(diff + ":" + n1Range + "," + n2Range + "," + Math.abs(intValue));
						} else if (diff > 0)
						{
							// if (intValue == 0)
							// {
							// if (verbosityLevelWhenReadingCM.intValue() > 0)
							// System.err.println("int value 0");
							//
							// totalReverseDRCorrect += (n1Range * n2Range) >> 1;
							// } else
							// {
							if (SymmetricDiscreteDistanceFunction.verbosityLevelWhenReadingCM.intValue() > 0)
							{
								System.err.println("add " + (diff >> 1) + "(" + n1Range + "x" + n2Range + ", intValue:" + intValue + ")");
							}

							totalReverseDRCorrect += diff >> 1;
							// }
						}
					}
					for (int i1 = node1.startIndex; i1 <= nodes[externalIndex1].endInternal; i1++)
					{
						for (int i2 = node2.startIndex; i2 <= nodes[externalIndex2].endInternal; i2++)
						{
							if (intValue > 0)
							{
								if (intValue > (Integer.MAX_VALUE >> 1))
								{
									increaseBlocks.add(new int[] { i1, i2 });
								}

								sym.set(i1, i2, 1);
								intValue--;

								if ((SymmetricDiscreteDistanceFunction.verbosityLevelWhenReadingCM.intValue() > 0)
										&& (((n1Range * n2Range) > 2) && (Math.abs(initialIntValue) != (n1Range * n2Range)))
										&& !(intValue >= (Integer.MAX_VALUE / 2)) && !(intValue <= (Integer.MIN_VALUE / 2)))
								{
									System.err.println("\tset edge (" + intValue + ")");
								}

							} else
							{
								if (intValue < (Integer.MIN_VALUE >> 1))
								{
									decreaseBlocks.add(new int[] { i1, i2 });
								}

								sym.set(i1, i2, 2);
								intValue++;

								if ((SymmetricDiscreteDistanceFunction.verbosityLevelWhenReadingCM.intValue() > 0)
										&& (((n1Range * n2Range) > 2) && (Math.abs(initialIntValue) != (n1Range * n2Range)))
										&& !(intValue >= (Integer.MAX_VALUE / 2)) && !(intValue <= (Integer.MIN_VALUE / 2)))
								{
									System.err.println("\tset non-edge (" + intValue + ")");
								}
							}
						}
					}
				}
			}
		}
		if (SymmetricDiscreteDistanceFunction.writerUncompressedCM_File.booleanValue())
		{
			final String pathname = "/home/sepphar/test.cm";
			SymmetricDiscreteDistanceFunction.writeUncompressedCMFile(sym, nodes, pathname);
		}
		reader.close();
		return new Store<Store<SymmetricDiscreteDistanceFunction, Integer>, Store<ArrayList<int[]>, ArrayList<int[]>>>(
				new Store<SymmetricDiscreteDistanceFunction, Integer>(sym, totalReverseDRCorrect), new Store<ArrayList<int[]>, ArrayList<int[]>>(
						increaseBlocks, decreaseBlocks));
	}

	public static int hashSumDist(final SymmetricDiscreteDistanceFunction dist)
	{
		return Arrays.deepHashCode(dist.distanceFunction);
	}

	public static int[] isConflict(final int i, final int j, final int l, final int d_i_j, final int d_i_l, final int d_j_l)
	{
		if ((d_i_j > d_i_l) && (d_i_j > d_j_l))
			return new int[] { Math.min(i, j), Math.max(i, j), l };
		else if ((d_i_l > d_i_j) && (d_i_l > d_j_l))
			return new int[] { Math.min(i, l), Math.max(i, l), j };
		else if ((d_j_l > d_i_j) && (d_j_l > d_i_l))
			return new int[] { Math.min(j, l), Math.max(j, l), i };
		return null;
	}

	public static int sumDist(final SymmetricDiscreteDistanceFunction dist)
	{
		int sum = 0;
		for (int i = 0; i < dist.getNoOfElements(); i++)
		{
			for (int j = i + 1; j < dist.getNoOfElements(); j++)
			{
				sum += dist.get(i, j) - 1;
			}
		}
		return sum;
	}

	public static void visualizeUltrametric(final String filePath, final SymmetricDiscreteDistanceFunction distanceFunction) throws IOException
	{
		class Cluster
		{

			int index;

			protected String name;

			public Cluster(final String i, final int index)
			{
				this.name = i;
				this.index = index;
			}
		}
		;
		class SuperCluster extends Cluster
		{

			private final Vector<Cluster> clusters = new Vector<Cluster>();

			public SuperCluster(final String name, final Cluster clu)
			{
				super(name, clu.index);
				this.clusters.add(clu);
			}

			public void add(final Cluster nextCluster)
			{
				this.clusters.add(nextCluster);
			}

			public void writeToFile(final FileWriter writer, final int distance) throws IOException
			{
				for (final Cluster i : this.clusters)
				{
					writer.write(this.name + " -- " + i.name + "\n");
				}
			}
		}
		;

		Vector<Cluster> clusters = new Vector<Cluster>();
		for (int i = 0; i < distanceFunction.getNoOfElements(); i++)
		{
			final Cluster clu = new Cluster(String.valueOf(i), i);
			clusters.add(clu);
		}
		final FileWriter writer = new FileWriter(new File(filePath));
		writer.write("graph G {\n");
		writer.write("ranksep = 2\n");

		int distance = 1;
		int count = 0;
		while (clusters.size() > 1)
		{
			final Vector<Cluster> newClusters = new Vector<Cluster>();
			while (!clusters.isEmpty())
			{
				final SuperCluster cluster = new SuperCluster("n" + count++, clusters.remove(0));

				final Iterator<Cluster> iter = clusters.iterator();
				while (iter.hasNext())
				{
					final Cluster nextCluster = iter.next();
					if (distanceFunction.get(cluster.index, nextCluster.index) == distance)
					{
						cluster.add(nextCluster);
						iter.remove();
					}
				}
				cluster.writeToFile(writer, distance);
				writer.flush();
				newClusters.add(cluster);
			}
			clusters = newClusters;
			distance++;
		}
		writer.write("}");
		writer.close();

	}

	public static void writeUncompressedCMFile(final SymmetricDiscreteDistanceFunction sym, final Node[] nodes, final String pathname)
			throws IOException
	{
		final FileWriter writer = new FileWriter(new File(pathname));
		writer.write(String.valueOf(sym.getNoOfElements()));
		for (final Node node : nodes)
		{
			for (int in = node.startIndex; in <= node.endInternal; in++)
			{
				writer.write("\n" + in);
			}
		}

		final Vector<Integer> indices = new Vector<Integer>();
		for (int i = 0; i < nodes.length; i++)
		{
			SymmetricDiscreteDistanceFunction.addAllIndices(indices, nodes[i]);
		}
		for (int nodeI = 0; nodeI < (indices.size() - 1); nodeI++)
		{
			writer.write("\n");
			for (int nodeJ = nodeI + 1; nodeJ < indices.size(); nodeJ++)
			{
				if (nodeJ > (nodeI + 1))
				{
					writer.write("\t");
				}
				if (sym.get(indices.get(nodeI), indices.get(nodeJ)) == 1)
				{
					writer.write("1");
				} else
				{
					writer.write("-1");
				}
			}
		}

		writer.close();
	}

	private final int[][] distanceFunction;

	private final boolean[] isDeletedData;

	private final int maxVal;

	private final int noElements;

	public SymmetricDiscreteDistanceFunction(final int n)
	{
		this(n, 0, -1);
	}

	/**
	 * Constructs a new distance function with randomized values.
	 * 
	 * @param n number of nodes
	 * @param maxDistValue range [1,...,maxDistValue]
	 */
	public SymmetricDiscreteDistanceFunction(final int n, final int maxDistValue, final int variation)
	{
		this(RandomDistanceFunctions.getRandomDistanceFunction(n, maxDistValue, variation, RandomDistanceFunctions.DEFAULT_MAX_DEGREE));
	}

	public SymmetricDiscreteDistanceFunction(final int[] nodes, final SymmetricDiscreteDistanceFunction dist)
	{
		this(nodes.length, 0, -1);
		for (int i = 0; i < nodes.length; i++)
		{
			for (int j = i + 1; j < nodes.length; j++)
			{
				this.set(i, j, dist.get(nodes[i], nodes[j]));
			}
		}
	}

	/**
	 * Constructs a new distance function.
	 * 
	 * @param distanceFunction
	 */
	public SymmetricDiscreteDistanceFunction(final int[][] distanceFunction)
	{
		this.distanceFunction = distanceFunction;
		this.isDeletedData = new boolean[this.distanceFunction.length + 1];
		Arrays.fill(this.isDeletedData, false);
		this.noElements = this.distanceFunction.length + 1;
		// traverseArray and Set Max Value
		int mv = 0;

		for (int i = 0; i < this.getNoOfElements(); i++)
		{
			for (int j = i + 1; j < this.getNoOfElements(); j++)
			{
				if (this.get(i, j) > mv)
				{
					mv = this.get(i, j);
				}
			}
		}
		this.maxVal = mv;
	}

	/**
	 * Copy constructor
	 * 
	 * @param function
	 */
	public SymmetricDiscreteDistanceFunction(final SymmetricDiscreteDistanceFunction function)
	{
		this(SymmetricDiscreteDistanceFunction.copyArray(function.distanceFunction));
	}

	protected void changeOccured(final int i, final int j, final int b, final int a)
	{
		this.setChanged();
		this.notifyObservers(new DistanceFunctionChange(i, j, b, a));
	}

	public void deleteData(final int index)
	{
		this.setDeleteStatus(index, true);
	}

	@Override
	public boolean equals(final Object obj)
	{
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (this.getClass() != obj.getClass())
			return false;
		final SymmetricDiscreteDistanceFunction other = (SymmetricDiscreteDistanceFunction) obj;
		if (!Arrays.equals(this.distanceFunction, other.distanceFunction))
			return false;
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction#get(int, int)
	 */
	@Override
	public int get(final int i, final int j)
	{
		if (j < i)
			return this.distanceFunction[j][i - (j + 1)];
		else
			return this.distanceFunction[i][j - (i + 1)];
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction#getDistance (distanceFunction.SymmetricDiscreteDistanceFunction)
	 */
	@Override
	public int getDistance(final SymmetricDiscreteDistanceFunction dis)
	{
		int dist = 0;
		for (int i = 0; i < this.getNoOfElements(); i++)
		{
			for (int j = i + 1; j < this.getNoOfElements(); j++)
			{
				dist += Math.abs(this.get(i, j) - dis.get(i, j));
			}
		}
		return dist;
	}

	public String getFileString(final String[] par, final String additionalHeader, final boolean writeHeader, final String sep,
			final String commentPrefix) throws IOException
	{
		// final FileWriter writer = new FileWriter(new File(filePath));
		final StringBuffer writer = new StringBuffer();
		// description

		if (writeHeader)
		{

			writer.append(commentPrefix + "n=" + this.getNoOfElements());
		}
		if (additionalHeader != null)
		{
			writer.append(commentPrefix + additionalHeader);
		}

		if (par != null)
		{
			for (int i = 0; (i < par.length) && ((i + 1) < par.length); i += 2)
			{
				writer.append("\t");
				writer.append(par[i] + "=" + par[i + 1]);
			}
		}
		for (final int[] element : this.distanceFunction)
		{
			writer.append("\n");
			for (int j = 0; j < element.length; j++)
			{
				if (j != 0)
				{
					writer.append(sep);
				}
				writer.append(String.valueOf(element[j]));
			}
		}
		return writer.toString();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction#getMaxValue ()
	 */
	@Override
	public int getMaxValue()
	{
		return this.maxVal;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction# getNoOfElements()
	 */
	@Override
	public int getNoOfElements()
	{
		return this.noElements;
	}

	public ReadOnlyViewOnSymmetricDiscreteDistanceFunction getUnmodifiableView()
	{
		class View implements ReadOnlyViewOnSymmetricDiscreteDistanceFunction
		{
			@Override
			public int get(final int i, final int j)
			{
				return SymmetricDiscreteDistanceFunction.this.get(i, j);
			}

			@Override
			public int getDistance(final SymmetricDiscreteDistanceFunction dis)
			{
				return SymmetricDiscreteDistanceFunction.this.getDistance(dis);
			}

			@Override
			public int getMaxValue()
			{
				return SymmetricDiscreteDistanceFunction.this.getMaxValue();
			}

			@Override
			public int getNoOfElements()
			{
				return SymmetricDiscreteDistanceFunction.this.getNoOfElements();
			}

			@Override
			public int[] isConflict(final int i, final int j, final int l)
			{
				return SymmetricDiscreteDistanceFunction.this.isConflict(i, j, l);
			}

			@Override
			public boolean isConflict(final int[] indices)
			{
				return SymmetricDiscreteDistanceFunction.this.isConflict(indices);
			}

			@Override
			public boolean isDeletedData(final int i)
			{
				return SymmetricDiscreteDistanceFunction.this.isDeletedData(i);
			}

			@Override
			public boolean isUltrametric()
			{
				return SymmetricDiscreteDistanceFunction.this.isUltrametric();
			}

			@Override
			public boolean isUltrametric(final Vector<Integer> indices)
			{
				return SymmetricDiscreteDistanceFunction.this.isUltrametric(indices);
			}
		}
		;
		return new View();
	}

	@Override
	public int hashCode()
	{
		return SymmetricDiscreteDistanceFunction.calcHash(this);
	}

	/**
	 * @param i
	 * @param j
	 * @param k
	 */
	public void increase(final int i, final int j, final int k)
	{
		this.set(i, j, this.get(i, j) + k);

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction#isConflict (int, int, int)
	 */
	@Override
	public int[] isConflict(final int i, final int j, final int l)
	{
		final int d_i_j = this.get(i, j);
		final int d_i_l = this.get(i, l);
		final int d_j_l = this.get(j, l);
		return SymmetricDiscreteDistanceFunction.isConflict(i, j, l, d_i_j, d_i_l, d_j_l);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction#isConflict (int[])
	 */
	@Override
	public boolean isConflict(final int[] indices)
	{
		final int[] arr = this.isConflict(indices[0], indices[1], indices[2]);
		if (arr != null)
		{
			indices[0] = arr[0];
			indices[1] = arr[1];
			indices[2] = arr[2];
			return true;
		}
		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction# isDeletedData(int)
	 */
	@Override
	public boolean isDeletedData(final int i)
	{
		return this.isDeletedData[i];
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction# isUltrametric()
	 */
	@Override
	public boolean isUltrametric()
	{
		for (int i = 0; i < this.getNoOfElements(); i++)
		{
			for (int j = i + 1; j < this.getNoOfElements(); j++)
			{
				for (int t = j + 1; t < this.getNoOfElements(); t++)
				{
					if (this.isConflict(i, j, t) != null)
						return false;
				}
			}
		}
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction# isUltrametric(java.util.Vector)
	 */
	@Override
	public boolean isUltrametric(final Vector<Integer> indices)
	{
		for (int i = 0; i < indices.size(); ++i)
		{
			for (int j = i + 1; j < indices.size(); ++j)
			{
				for (int k = j + 1; k < indices.size(); ++k)
				{
					if (this.isConflict(indices.get(i), indices.get(j), indices.get(k)) != null)
						return false;
				}
			}
		}
		return true;
	}

	/**
	 * @param i
	 * @param j
	 * @param k
	 */
	public void set(final int i, final int j, final int k)
	{
		final int before = this.get(i, j);
		if (j < i)
		{
			this.distanceFunction[j][i - (j + 1)] = k;
		} else
		{
			this.distanceFunction[i][j - (i + 1)] = k;
		}
		final int after = this.get(i, j);
		this.changeOccured(i, j, before, after);
	}

	public void set(final SymmetricDiscreteDistanceFunction lpDistFunction)
	{
		if (!(lpDistFunction.getNoOfElements() == this.getNoOfElements()))
			throw new IllegalArgumentException();
		for (int i = 0; i < this.getNoOfElements(); i++)
		{
			for (int j = i + 1; j < this.getNoOfElements(); j++)
			{
				this.set(i, j, lpDistFunction.get(i, j));
			}
		}
	}

	public void setDeleteStatus(final int index, final boolean del)
	{
		this.isDeletedData[index] = del;
	}

	@Override
	public String toString()
	{
		final StringBuffer s = new StringBuffer();
		for (int i = 0; i <= this.distanceFunction.length; i++)
		{
			s.append(i + "\t");
		}
		s.append('\n');
		for (int i = 0; i <= this.distanceFunction.length; i++)
		{
			s.append("_______");
		}
		s.append('\n');
		for (int i = 0; i < this.distanceFunction.length; i++)
		{
			final int[] arr = this.distanceFunction[i];
			s.append(i + "|\t");
			for (int t = 0; t < i; t++)
			{
				s.append("\t");
			}
			for (int j = 0; j < arr.length; j++)
			{
				if (j == 0)
				{
					s.append("{");
				}
				s.append(arr[j]);

				if (j < (arr.length - 1))
				{
					s.append(",");
					s.append("\t");
				} else
				{
					s.append("},");
				}
			}
			s.append('\n');
		}
		s.append('\n');
		return s.toString();
	}

	public void writeToFile(final String filePath, final String additionalHeader) throws IOException
	{
		this.writeToFile(filePath, null, additionalHeader);
	}

	public void writeToFile(final String filePath, final String[] par, final String additionalHeader) throws IOException
	{

		final FileWriter writer = new FileWriter(new File(filePath));
		writer.write(this.getFileString(par, additionalHeader, true, SymmetricDiscreteDistanceFunction.sep, "# "));
		writer.close();
	}
}