/**
 * 
 */
package distanceFunction.random;

import java.util.Arrays;
import java.util.Random;
import java.util.Vector;

/**
 * This class provides the methods which can be used to create random distance functions.
 * 
 * @author Sepp Hartung
 * @version 1.1
 */
public class RandomDistanceFunctions
{
	static class Range
	{
		private final int i;

		private final int j;

		public Range(final int i, final int j)
		{
			this.i = i;
			this.j = j;
		}

		private int getLeftMargin()
		{
			return this.i;
		}

		private int getRightMargin()
		{
			return this.j;
		}

		@Override
		public String toString()
		{
			return "(" + this.getLeftMargin() + "," + this.getRightMargin() + ")";
		}
	}

	public static final int DEFAULT_MAX_DEGREE = 5;

	private static Range[] arrangeLevel(final Range range, final int value, int maxSplit, int minSplit, final int[][] distance)
	{
		final Random ran = new Random();
		final int leftMargin = range.getLeftMargin();
		final int rightMargin = range.getRightMargin();

		if (rightMargin > leftMargin)
		{
			final int abs = rightMargin - leftMargin;
			maxSplit = Math.min(maxSplit, abs);
			minSplit = Math.min(minSplit, maxSplit);

			final int degree = minSplit + ran.nextInt((maxSplit + 1) - minSplit);
			if (degree > 1)
			{
				final int[] margins = new int[degree - 1];
				boolean flag = true;
				while (flag)
				{
					for (int i = 0; i < margins.length; i++)
					{
						margins[i] = ran.nextInt(abs);
					}
					Arrays.sort(margins);
					flag = false;
					for (int i = 0; i < (margins.length - 1); i++)
					{
						flag = flag || (margins[i] == margins[i + 1]);
					}
				}
				final Range[] nodes = new Range[degree];
				for (int i = 0; i < nodes.length; i++)
				{
					final int leMar = i > 0 ? margins[i - 1] + 1 : 0;
					final int riMar = (i < (nodes.length - 1)) && (degree > 1) ? margins[i] : abs;
					nodes[i] = new Range(leMar + leftMargin, riMar + leftMargin);
				}
				RandomDistanceFunctions.setDistance(nodes, value, distance);
				return nodes;
			} else
			{
				final Range[] nodes = new Range[1];
				nodes[0] = range;
				return nodes;
			}
		}
		return null;

	}

	/**
	 * Derange the given distance function (uniform distribution).
	 * 
	 * @param distance distance function
	 * @param maxDistValue range of distance values
	 * @param variations number of places which will be deranged
	 * @return
	 */
	public static int[][] derangedUltrametric(final int[][] distance, final int maxDistValue, final int variations)
	{
		final int n = distance.length + 1;
		final int[][] tmp = RandomDistanceFunctions.getDistArr(n);
		// derange distances
		int cDerange = variations;
		final Random random = new Random();
		while (cDerange > 0)
		{
			int i = random.nextInt(n);
			int j = random.nextInt(n - 1);
			if (i > j)
			{
				final int t = i;
				i = j;
				j = t;
			}
			if (i == j)
			{
				j++;
			}
			j = j - (i + 1);
			final int correct = random.nextFloat() <= 0.5f ? 1 : -1;
			if ((correct == 1) && (distance[i][j] < maxDistValue) && (tmp[i][j] >= 0))
			{
				distance[i][j]++;
				tmp[i][j]++;
				cDerange--;
			} else if ((distance[i][j] > 1) && (tmp[i][j] <= 0))
			{
				distance[i][j]--;
				tmp[i][j]--;
				cDerange--;
			}
		}
		return distance;
	};

	private static int[][] getDistArr(final int n)
	{
		final int[][] distance = new int[n - 1][];
		for (int i = n - 1; i > 0; i--)
		{
			distance[n - 1 - i] = new int[i];
		}
		return distance;
	}

	/**
	 * Returns a random distance functions which is build by a derangement of an ultrametric.
	 * 
	 * @param n number of nodes
	 * @param maxDistValue
	 * @param variations
	 * @return
	 */
	public static int[][] getRandomDistanceFunction(int n, int maxDistValue, final int variations, final int maxDegree)
	{
		n = n >= 3 ? n : 3;
		if (maxDistValue < 0)
		{
			maxDistValue = (int) Math.round(Math.log(n + 2));
		}
		maxDistValue = Math.max(maxDistValue, 0);

		if (maxDistValue > 0)
		{// ini with random distances
			// create an ultrametric
			final int[][] distance = RandomDistanceFunctions.getUltrametric(n, maxDistValue, maxDegree);
			return RandomDistanceFunctions.derangedUltrametric(distance, maxDistValue, variations);
		} else
		{// empty distance function
			final int[][] distance = RandomDistanceFunctions.getDistArr(n);
			return distance;
		}
	}

	/**
	 * Returns a random ultrametric.
	 * 
	 * @param n number of nodes
	 * @param maxDistValue distance value range [1..maxDistValue]
	 * @return
	 */
	public static int[][] getUltrametric(final int n, final int maxDistValue, final int maxSplit)
	{
		final int[][] distance = RandomDistanceFunctions.getDistArr(n);

		Range[] ranges = { new Range(0, n - 1) };
		int value = maxDistValue;
		while (value > 0)
		{
			final Vector<Range> tmp = new Vector<Range>();
			for (final Range range : ranges)
			{
				final Range[] t = RandomDistanceFunctions.arrangeLevel(range, value, maxSplit, value == maxDistValue ? 2 : 1, distance);
				if (t != null)
				{
					for (final Range r : t)
					{
						tmp.add(r);
					}
				}
			}
			ranges = tmp.toArray(new Range[0]);
			value--;
		}
		for (final Range range : ranges)
		{
			for (int i = range.getLeftMargin(); i < range.getRightMargin(); i++)
			{
				for (int j = i + 1; j <= range.getRightMargin(); j++)
				{
					RandomDistanceFunctions.set(i, j, 1, distance);
				}
			}
		}
		return distance;
	}

	private static void set(final int i, final int j, final int k, final int[][] distance)
	{
		if (j < i)
		{
			distance[j][i - (j + 1)] = k;
		} else
		{
			distance[i][j - (i + 1)] = k;
		}
	}

	private static void setDistance(final Range[] nodes, final int value, final int[][] distance)
	{
		for (int i = 0; i < nodes.length; i++)
		{
			for (int j = i + 1; j < nodes.length; j++)
			{
				for (int t = nodes[i].getLeftMargin(); t <= nodes[i].getRightMargin(); t++)
				{
					for (int s = nodes[j].getLeftMargin(); s <= nodes[j].getRightMargin(); s++)
					{
						RandomDistanceFunctions.set(t, s, value, distance);
					}
				}
			}
		}
	}
}
