package dataReductionRules;

import java.util.PriorityQueue;
import java.util.Vector;

import solver.hierarchicalClustering.SearchTreeSolver;
import solver.hierarchicalClustering.SolverInterface;
import dataReductionRules.ChangeLogger.DistanceChange;
import distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction;
import distanceFunction.SymmetricDiscreteDistanceFunction;

/**
 * 
 * (k+1)-rule (lazy version)o
 * 
 * @author Sepp Hartung
 * @version 1.0
 * 
 */
public class Rule0 extends AdvancedRule implements Cloneable
{
	public static final String ruleKey = "lazy k+1 rule";

	public static Vector<int[]> getIndependentConflicts(final SolverInterface solver)
	{
		final ReadOnlyViewOnSymmetricDiscreteDistanceFunction conflictEdge = solver.getConflictEdgeStatistic();
		final ReadOnlyViewOnSymmetricDiscreteDistanceFunction nonConflictEdge = solver.getNonConflictEdgeStatistic();

		final SymmetricDiscreteDistanceFunction distanceFunction = solver.getDistanceFunction();

		final Vector<int[]> independentConflicts = new Vector<int[]>(Math.max(10, solver.getParameterValue()));
		{
			final EdgeContainerInterface edges = new Rule1.EdgeContainer2(distanceFunction.getNoOfElements());
			final PriorityQueue<ConflictVal> queue = Rule1.iniConflictQueue(solver.getConflicts(), conflictEdge, nonConflictEdge, edges,
					Integer.MAX_VALUE).getS2();

			while (!queue.isEmpty())
			{
				final int[] conflict = queue.poll().getValue();
				if (Rule1.areConflictEdgesDisjointToKnown(edges, conflict))
				{
					edges.addConflict(conflict);
					independentConflicts.add(conflict);
				}
			}
		}
		return independentConflicts;
	}

	private static int getMaxIndependentConflicts(final SymmetricDiscreteDistanceFunction distanceFunction, final int i, final int j,
			final Vector<int[]> independentConflicts)
	{
		int lowerBound = independentConflicts.size();
		// final int c = lowerBound;
		for (final int[] conflict : independentConflicts)
		{
			if (Rule0.isConflictContainingEdge(i, j, conflict))
			{
				lowerBound--;
			} else
			{
				for (final int l : conflict)
				{
					if ((i != l) && (j != l) && (Rule0.isConflictContainingEdge(i, l, conflict) || Rule0.isConflictContainingEdge(j, l, conflict))
							&& (distanceFunction.isConflict(i, j, l) != null))
					{
						lowerBound--;
					}
				}
			}
		}
		// System.out.println("lowerbound decrease:"+(c-lowerBound));

		return lowerBound;
	}

	public static boolean isConflictContainingEdge(final int i, final int j, final int[] conflict)
	{
		if (conflict[0] == i)
			return (conflict[1] == j) || (conflict[2] == j);
		else if (conflict[1] == i)
			return (conflict[0] == j) || (conflict[2] == j);
		else if (conflict[2] == i)
			return (conflict[0] == j) || (conflict[1] == j);
		return false;
	}

	public Rule0(final int interleavingConst)
	{
		super(interleavingConst);
	}

	@Override
	public boolean performReductionRule(final ChangeLogger solver) throws UnsolvableException
	{
		boolean reducedSomething = false;

		final ReadOnlyViewOnSymmetricDiscreteDistanceFunction conflictEdge = solver.getConflictEdgeStatistic();
		final ReadOnlyViewOnSymmetricDiscreteDistanceFunction nonConflictEdge = solver.getNonConflictEdgeStatistic();

		final SymmetricDiscreteDistanceFunction distanceFunction = solver.getDistanceFunction();

		final Vector<int[]> independentConflicts = Rule0.getIndependentConflicts(solver);

		if (independentConflicts.size() > solver.getParameterValue())
			throw new UnsolvableException();

		if (SearchTreeSolver.RulesVerboseLevel > 1)
		{
			System.out.println("lazy lowerbound:" + independentConflicts.size());
		}

		final Vector<ChangeLogger.DistanceChange> changes = new Vector<>();

		for (int i = 0; i < distanceFunction.getNoOfElements(); i++)
		{
			for (int j = i + 1; j < distanceFunction.getNoOfElements(); j++)
			{
				final int lowerBound = Rule0.getMaxIndependentConflicts(distanceFunction, i, j, independentConflicts);

				{
					final boolean greater = (conflictEdge.get(i, j) + lowerBound) > solver.getParameterValue();
					final boolean less = (nonConflictEdge.get(i, j) + lowerBound) > solver.getParameterValue();
					// System.out.println(conflictEdge.get(i, j) + "|" +
					// nonConflictEdge.get(i, j));

					if ((greater && less) || ((greater || less) && (solver.getParameterValue() == 0)))
						// System.out.println("unsolvable " + i + "|" +
						// j+",lowerbound:"+lowerBound);
						throw new UnsolvableException();
					else if (greater || less)
					{
						if (greater)
						{// decrease d_i_j
							changes.add(new DistanceChange(i, j, -1, true));
							reducedSomething = true;
						} else
						{// increase
							changes.add(new DistanceChange(i, j, 1, true));
							reducedSomething = true;
							// System.out.println("increase " + i + "|" +
							// j+",lowerbound:"+lowerBound);
						}
						if ((SearchTreeSolver.RulesVerboseLevel > 1) && ((changes.size() % 100) == 0))
						{
							System.out.println("number of already saved changes:" + changes.size());
						}
					}
				}
			}
		}
		if (SearchTreeSolver.RulesVerboseLevel > 1)
		{
			System.out.println("changes:" + changes.size());
		}

		for (final DistanceChange change : changes)
		{
			solver.changeDistanceValue(change.getI(), change.getJ(), change.getC(), true);
		}
		// System.out.println("performed");
		// System.out.println(solver.getParameterValue() + "|" +
		// solver.getConflicts().size());
		return reducedSomething;
	}
}
