package dataReductionRules.rule3HelpClasses;

import java.util.Iterator;

/**
 * 
 * @author Tomasz Przedmojski
 * 
 */
public class CliqueGraph implements Iterable<Clique>
{
	private final int maxClique;
	private final int maxNode;

	private SparseIntegerIdenfitiableSet<Clique>[] neighborhood;
	private Clique[] nodeCliqueMapping;

	private final CliquePool pool;

	private final int RESERVED = 2;
	private final ThresholdGraph thresholdGraph;

	/**
	 * Constructs CliqueGraph from passed ThresholdGraph. CliqueGraph is different representation for ThresholdGraph in which all nodes with the same
	 * neighborhood are joined and stored as one node.
	 * 
	 * Resulting graph is more sparse.
	 * 
	 * @param thresholdGraph source graph
	 */
	public CliqueGraph(final ThresholdGraph thresholdGraph) throws CliquePoolExhausted
	{
		this.thresholdGraph = thresholdGraph;
		this.maxNode = thresholdGraph.getMaxNodes();
		this.maxClique = this.maxNode + this.RESERVED;

		this.pool = new CliquePool(this.maxClique, this.maxNode);

		this.constructCliques();
		this.constructGraph();
		this.maximizeCliques();
	}

	/**
	 * Joins cliques a and b
	 * 
	 * @param absorber the clique that "absorbs"
	 * @param absorbee the clique that is "being absorbed"
	 */
	private void absorbClique(final Clique absorber, final Clique absorbee)
	{
		// move all elements from clique b to clique a
		this.moveCliqueElements(absorbee, absorber);
		// remove clique b from graph
		this.removeCliqueFromGraph(absorbee);
	}

	/**
	 * Returns new clique containing only passed node.
	 * 
	 * @param thresholdGraphNode node from threshold graph
	 * @return instance of Clique
	 */
	protected Clique addCliqueToGraph(final int thresholdGraphNode) throws CliquePoolExhausted
	{
		// create clique
		final Clique newClique = this.pool.getClique();
		// add starting element to clique
		newClique.add(thresholdGraphNode);
		// store association with threshond graph node
		this.nodeCliqueMapping[thresholdGraphNode] = newClique;

		assert newClique.getNumber() < this.maxClique;

		// return the clique
		return newClique;
	}

	public void addEdge(final Clique source, final Clique target)
	{
		this.neighborhood[source.getNumber()].add(target);
		this.neighborhood[target.getNumber()].add(source);
	}

	/**
	 * Constructs a mapping between nodes and containing cliques.
	 */
	private void constructCliques() throws CliquePoolExhausted
	{
		this.nodeCliqueMapping = new Clique[this.maxNode];
		for (int u = 0; u < this.maxNode; ++u)
		{
			this.nodeCliqueMapping[u] = this.addCliqueToGraph(u);
		}
	}

	/**
	 * Constructs cc graph, basically a copy of the threshold graph
	 */
	private void constructGraph()
	{
		this.neighborhood = new SparseIntegerIdenfitiableSet[this.maxClique];

		for (int i = 0; i < this.maxClique; ++i)
		{
			this.neighborhood[i] = new SparseIntegerIdenfitiableSet<Clique>(this.maxClique);
		}

		for (int i = 0; i < this.nodeCliqueMapping.length; ++i)
		{
			for (int j = i + 1; j < this.nodeCliqueMapping.length; ++j)
			{
				if (this.thresholdGraph.containsEdge(i, j))
				{
					this.addEdge(this.nodeCliqueMapping[i], this.nodeCliqueMapping[j]);
				}
			}
		}
	}

	public boolean containsEdge(final Clique source, final Clique target)
	{
		return this.neighborhood[source.getNumber()].contains(target);
	}

	/**
	 * Copies all neighbors from sourceClique to targetClique. After this operation N(targetClique) will be a superset of N(sourceClique).
	 * 
	 * @param sourceClique
	 * @param targetClique
	 */
	private void copyCliqueNeighborhood(final Clique sourceClique, final Clique targetClique)
	{
		for (final Clique neighborClique : this.neighborsOf(sourceClique))
		{
			this.addEdge(neighborClique, targetClique);
		}

	}

	/**
	 * Repairs graph structure after an edge has been added to the threshold graph.
	 * 
	 * @param i
	 * @param j
	 */
	public void edgeInThresholdGraphAdded(final int i, final int j) throws CliquePoolExhausted
	{
		final Clique cliqueI = this.nodeCliqueMapping[i];
		final Clique cliqueJ = this.nodeCliqueMapping[j];

		if (cliqueI == cliqueJ)
		{
			// i and j already are in the same clique - no change occured!
			assert false;
			return;
		}

		final Clique newIClique = this.extractFromClique(cliqueI, i);
		final Clique newJClique = this.extractFromClique(cliqueJ, j);

		this.addEdge(newIClique, newJClique);

		if (!cliqueI.isEmpty())
		{
			this.addEdge(cliqueI, newIClique);
		} else
		{
			this.removeCliqueFromGraph(cliqueI);
		}
		if (!cliqueJ.isEmpty())
		{
			this.addEdge(cliqueJ, newJClique);
		} else
		{
			this.removeCliqueFromGraph(cliqueJ);
		}

		if (this.isCliquePresent(cliqueI))
		{
			this.maximizeClique(cliqueI);
		}
		if (this.isCliquePresent(newIClique))
		{
			this.maximizeClique(newIClique);
		}
		if (this.isCliquePresent(cliqueJ))
		{
			this.maximizeClique(cliqueJ);
		}
		if (this.isCliquePresent(newJClique))
		{
			this.maximizeClique(newJClique);
		}
	}

	/**
	 * Repairs graph structure after an edge has been removed from the threshold graph.
	 * 
	 * @param i
	 * @param j
	 */
	public void edgeInThresholdGraphRemoved(final int i, final int j) throws CliquePoolExhausted
	{
		final Clique cliqueI = this.nodeCliqueMapping[i];
		final Clique cliqueJ = this.nodeCliqueMapping[j];

		assert cliqueI != null;
		assert cliqueJ != null;

		if (cliqueI == cliqueJ)
		{
			final Clique newIClique = this.extractFromClique(cliqueI, i);
			final Clique newJClique = this.extractFromClique(cliqueJ, j);
			// This should not be neccessary
			// graph.removeEdge(newIClique.getNumber(), newJClique.getNumber());

			if (!cliqueI.isEmpty())
			{
				this.addEdge(cliqueI, newIClique);
			} else
			{
				this.removeCliqueFromGraph(cliqueI);
			}
			if (this.isCliquePresent(cliqueI))
			{
				this.maximizeClique(cliqueI);
			}
			if (this.isCliquePresent(newIClique))
			{
				this.maximizeClique(newIClique);
			}
			if (this.isCliquePresent(newJClique))
			{
				this.maximizeClique(newJClique);
			}
		} else
		{
			if (this.containsEdge(cliqueI, cliqueJ))
			{
				final Clique newIClique = this.extractFromClique(cliqueI, i);
				final Clique newJClique = this.extractFromClique(cliqueJ, j);
				// This should not be neccessary
				// graph.removeEdge(newIClique.getNumber(),
				// newJClique.getNumber());

				if (!cliqueI.isEmpty())
				{
					this.addEdge(cliqueI, newIClique);
				} else
				{
					this.removeCliqueFromGraph(cliqueI);
				}
				if (!cliqueJ.isEmpty())
				{
					this.addEdge(cliqueJ, newJClique);
				} else
				{
					this.removeCliqueFromGraph(cliqueJ);
				}

				this.removeEdge(newIClique, newJClique);

				if (this.isCliquePresent(cliqueI))
				{
					this.maximizeClique(cliqueI);
				}
				if (this.isCliquePresent(newIClique))
				{
					this.maximizeClique(newIClique);
				}
				if (this.isCliquePresent(cliqueJ))
				{
					this.maximizeClique(cliqueJ);
				}
				if (this.isCliquePresent(newJClique))
				{
					this.maximizeClique(newJClique);
				}
			}
		}
	}

	/**
	 * Checks if nodes from cliques a and b have equal neighbors in threshold graph
	 * 
	 * @param cliqueA
	 * @param cliqueB
	 * @return true if cliques are equivalent
	 */
	private boolean equalNeighborhoods(final Clique cliqueA, final Clique cliqueB)
	{
		return this.thresholdGraph.neighborsOf(cliqueA.any()).equals(this.thresholdGraph.neighborsOf(cliqueB.any()));
	}

	/**
	 * Removes node from passed clique. Registers new clique and adds this node to the clique. Copies neighborhood from passed clique - new clique
	 * will have all the same neighbors in graph.
	 * 
	 * @param clique
	 * @param node
	 * @return new instance of Clique containing node
	 */
	private Clique extractFromClique(final Clique clique, final int node) throws CliquePoolExhausted
	{
		final Clique newNodeClique = this.addCliqueToGraph(node);
		clique.remove(node);
		this.copyCliqueNeighborhood(clique, newNodeClique);
		this.nodeCliqueMapping[node] = newNodeClique;

		return newNodeClique;
	}

	/***********************************************
	 * 
	 * ******************************************** Callback methods for changes in source graph ********************************************
	 * 
	 ***********************************************/

	public int getCliqueCount()
	{
		return this.pool.size();
	}

	public Clique getCliqueOfNodeInThresholdGraph(final int nodeInThresholdGraph)
	{
		return this.nodeCliqueMapping[nodeInThresholdGraph];
	}

	/***********************************************/
	/***********************************************/
	/***********************************************/

	public boolean isCliquePresent(final Clique clique)
	{
		return this.pool.getSet().contains(clique);
	}

	@Override
	public Iterator<Clique> iterator()
	{
		return this.pool.getSet().iterator();
	}

	/**
	 * Maximizes given clique iterates over neighboring cliques and "absorbs" them if they have equal neighborhood.
	 * 
	 * @param clique clique
	 * @return true if changes (absorptions) have been performed, false otherwise
	 */
	protected boolean maximizeClique(final Clique clique)
	{
		final SparseIntegerIdenfitiableSet<Clique> set = this.neighborsOf(clique);
		final Object[] array = set.getDenseArray();

		int index = 0;
		boolean changed = false;

		while (index < set.size())
		{
			final Clique neighbor = (Clique) array[index];

			if (this.equalNeighborhoods(clique, neighbor))
			{
				this.absorbClique(clique, neighbor);
				changed = true;
			} else
			{
				++index;
			}
		}

		return changed;
	}

	/**
	 * Ensures that all cliques are maximal
	 */
	private void maximizeCliques()
	{
		for (final Clique clique : this.pool.getSet())
		{
			if (this.maximizeClique(clique))
			{
				// since there has been some changes it is neccessary
				// to restart the process maximization process
				this.maximizeCliques();
				return;
			}
		}
	}

	/**
	 * Moves all elements from source clique to target clique
	 * 
	 * @param source source clique
	 * @param target target clique
	 */
	private void moveCliqueElements(final Clique source, final Clique target)
	{
		for (final int node : source)
		{
			target.add(node);
			this.nodeCliqueMapping[node] = target;
		}
		source.clear();
	}

	public int neighborCount(final Clique clique)
	{
		return this.neighborhood[clique.getNumber()].size();
	}

	public SparseIntegerIdenfitiableSet<Clique> neighborsOf(final Clique clique)
	{
		return this.neighborhood[clique.getNumber()];
	}

	/**
	 * Removes given clique from graph.
	 * 
	 * @param clique
	 */
	protected void removeCliqueFromGraph(final Clique clique)
	{
		for (final int thresholdGraphNode : clique)
		{
			this.nodeCliqueMapping[thresholdGraphNode] = null;
		}

		while (!this.neighborsOf(clique).isEmpty())
		{
			this.removeEdge(clique, (Clique) this.neighborsOf(clique).getDenseArray()[0]);
		}

		for (final Clique any : this.pool.getSet())
		{
			assert !this.containsEdge(any, clique);
			assert !this.containsEdge(clique, any);
		}

		this.pool.freeClique(clique);

	}

	public void removeEdge(final Clique source, final Clique target)
	{
		this.neighborhood[source.getNumber()].remove(target);
		this.neighborhood[target.getNumber()].remove(source);
	}

}
