package dataReductionRules.rule3HelpClasses;

import java.util.Iterator;

import dataReductionRules.rule3HelpClasses.hashing.SummingHashingAlgorithm;

/**
 * Integer set representation suitable for sparse integer sets bounded by low integers. Really nice upper bounds on most methods.
 * 
 * @author Tomasz Przedmojski
 * 
 */
public class SparseIntegerSet implements Iterable<Integer>
{

	protected int dense[];
	protected final SummingHashingAlgorithm hashingAlgorithm;
	protected int members;
	protected int sparse[];

	/**
	 * Constructs the set. Performance guarantee: O(n)
	 * 
	 * @param maxValue (n) maximum possible value in set.
	 */
	public SparseIntegerSet(final int maxValue)
	{
		this.sparse = new int[maxValue + 1];
		this.dense = new int[maxValue + 1];
		for (int i = 0; i <= maxValue; ++i)
		{
			this.dense[i] = -1;
		}
		this.members = 0;

		this.hashingAlgorithm = new SummingHashingAlgorithm();
	}

	/**
	 * Adds element with given value to the set. Performance guarantee: O(1)
	 * 
	 * @param value value of the element
	 * @return true if this set did not already contain the specified element.
	 */
	public boolean add(final int value)
	{
		if (!this.contains(value))
		{
			this.sparse[value] = this.members;
			this.dense[this.members] = value;
			this.members++;

			this.hashingAlgorithm.elementAdded(value);
			return true;
		}
		return false;
	}

	/**
	 * Removes all elements from the set. Performance guarantee: O(1)
	 */
	public void clear()
	{
		this.members = 0;
	}

	/**
	 * Checks if element with given value is present in set. Performance guarantee: O(1)
	 * 
	 * @param value value of the element
	 * @return true if element is present, false otherwise.
	 */
	public boolean contains(final int value)
	{
		final int k = this.sparse[value];
		return (k < this.members) && (this.dense[k] == value);
	}

	@Override
	public boolean equals(final Object obj)
	{
		if (obj == null)
			return false;
		if (this.getClass() != obj.getClass())
			return false;
		final SparseIntegerSet other = (SparseIntegerSet) obj;

		if (this.hashCode() != other.hashCode())
			return false;

		if (this.members != other.members)
			return false;

		for (int i = 0; i < this.members; ++i)
		{
			if (!other.contains(this.dense[i]))
				return false;
		}

		return true;
	}

	/**
	 * Returns the dense array of the set. Together with size() it can be used to iterate over all elements in the set in O(n). Performance guarantee
	 * for iterating over the elements: O(n)
	 * 
	 * @return
	 */
	public int[] getDenseArray()
	{
		return this.dense;
	}

	@Override
	public int hashCode()
	{
		// int hash = 7;
		// hash = 13 * hash + this.members;
		// hash = 13 * hash + Arrays.hashCode(this.sparse);
		// hash = 13 * hash + Arrays.hashCode(this.dense);
		// return hash;
		return this.hashingAlgorithm.getHashCode();
	}

	/**
	 * Checks if any elements are present in set. Performance guarantee: O(1)
	 * 
	 * @return true if set is not empty, false otherwise.
	 */
	public boolean isEmpty()
	{
		return this.members == 0;
	}

	@Override
	public Iterator<Integer> iterator()
	{
		return new Iterator<Integer>()
		{
			private int index = 0;

			@Override
			public boolean hasNext()
			{
				return this.index < SparseIntegerSet.this.size();
			}

			@Override
			public Integer next()
			{
				return SparseIntegerSet.this.dense[this.index++];
			}

			@Override
			public void remove()
			{
				throw new UnsupportedOperationException("Removing elements is not supported.");
			}
		};
	}

	/**
	 * Removes element with given value from the set. For elements not present in the set this method is idempotent. Performance guarantee: O(1)
	 * 
	 * @param value value of the element
	 * @return true if this set did contain the specified element.
	 */
	public boolean remove(final int value)
	{
		if (!this.contains(value))
			return false;
		else
		{
			this.hashingAlgorithm.elementRemoved(value);
		}

		if ((this.members == 1) || (this.sparse[value] == this.members))
		{
			this.members--;
		} else
		{
			final int freedIndex = this.sparse[value];
			final int newValue = this.dense[this.members - 1];

			this.sparse[newValue] = freedIndex;
			this.dense[freedIndex] = newValue;
			this.members--;
		}
		return true;
	}

	/**
	 * Returns the number of elements in the set. Use in conjuction with getDenseArray() Performance guarantee: O(1)
	 */
	public int size()
	{
		return this.members;
	}

}