package solver.heuristics;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;

import org.apache.commons.lang3.mutable.MutableDouble;
import org.apache.commons.lang3.mutable.MutableInt;

import solver.hierarchicalClustering.Conflicts;
import solver.hierarchicalClustering.HierarchicalClusteringSolver;
import sun.reflect.generics.reflectiveObjects.NotImplementedException;
import ProgOpt.SmacClusterEditOpt;
import benchmark.helpClasses.StopWatch;
import dataReductionRules.ConflictVal;
import dataReductionRules.Rule1;
import dataReductionRules.UnsolvableException;
import dataStructures.utils.ArrayS;
import distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction;
import distanceFunction.SymmetricBooleanArrayDistanceFunction;
import distanceFunction.SymmetricDiscreteDistanceFunction;

public class GreedyLowerBoundHeuristic extends HierarchicalClusteringSolver
{
	public static final MutableInt GreedyLowerBound_modShuffle = new MutableInt(6);

	public static final MutableInt heuristicLowerBoundRuns = new MutableInt(200);

	public static final MutableDouble highestShuffleFrac = new MutableDouble(0.2);

	public static final MutableDouble lowerBoundTimeProportion = new MutableDouble(0.03d);

	/**
	 * @param conflictVals
	 * @param maxKey
	 */
	private static void sortConflicts(final ArrayList<ConflictVal> conflictVals, final int maxKey)
	{
		// if (maxKey < 10000)
		// {
		// ArrayS.bucketSort(conflictVals, maxKey + 1);
		// } else
		{
			Collections.sort(conflictVals);
		}
	}

	public GreedyLowerBoundHeuristic(final SymmetricDiscreteDistanceFunction dist)
	{
		super(dist, -1);
	}

	public GreedyLowerBoundHeuristic(final SymmetricDiscreteDistanceFunction dist, final Conflicts conflicts)
	{
		super(dist, -1, conflicts);
	}

	public int getLowerBound()
	{
		final double maxTime = GreedyLowerBoundHeuristic.lowerBoundTimeProportion.doubleValue() * SmacClusterEditOpt.cutoff_time.doubleValue();
		return this.getLowerBound(maxTime, GreedyLowerBoundHeuristic.highestShuffleFrac.floatValue(), 0f,
				GreedyLowerBoundHeuristic.heuristicLowerBoundRuns.intValue());
	}

	public int getLowerBound(final double maxTime, final float highShuffleFrac, final float minShuffleFrac, final int numberOfRounds)
	{
		if ((highShuffleFrac > 1f) || (minShuffleFrac > highShuffleFrac) || (minShuffleFrac < 0f))
			throw new IllegalAccessError();

		// System.err.println("start lower bound:" + maxTime + ", " + highShuffleFrac + ", " + minShuffleFrac + ", " + numberOfRounds);

		final ArrayList<ConflictVal> conflictVals;
		int maxKey = Integer.MIN_VALUE;

		{// add and sort all conflicts
			final Collection<int[]> conflicts = this.getConflictStorage().getConflicts();
			conflictVals = new ArrayList<>(conflicts.size());
			{
				final ReadOnlyViewOnSymmetricDiscreteDistanceFunction conflictEdge = this.getConflictStorage().getConflictEdgeStatistic();
				final ReadOnlyViewOnSymmetricDiscreteDistanceFunction nonConflictEdge = this.getConflictStorage().getNonConflictEdgeStatistic();
				for (final int[] con : conflicts)
				{
					final int key = Rule1.getConflictCount(conflictEdge, nonConflictEdge, con);
					maxKey = Math.max(key, maxKey);
					conflictVals.add(new ConflictVal(key, con));
				}

			}
			GreedyLowerBoundHeuristic.sortConflicts(conflictVals, maxKey);
		}

		int lowerBoundGlobal = Integer.MIN_VALUE;
		final StopWatch watch = new StopWatch();

		boolean sorted = true;

		final int modShuffle = GreedyLowerBoundHeuristic.GreedyLowerBound_modShuffle.intValue();

		final float step = (float) ((highShuffleFrac - minShuffleFrac) / Math.ceil(numberOfRounds / (double) modShuffle));

		for (int rounds = 1; rounds <= numberOfRounds; rounds++)
		{
			if (!sorted)
			{
				GreedyLowerBoundHeuristic.sortConflicts(conflictVals, maxKey);
				sorted = true;
			}
			ArrayS.shuffleEqualSortKeyElements(conflictVals);
			if ((rounds % modShuffle) == 2)
			{
				final float per = ((float) Math.ceil(rounds / (float) modShuffle) * step) + minShuffleFrac;
				ArrayS.shuffleWithinSmallRange(conflictVals, per);
				sorted = false;
				// System.err.println("\tshuffle:" + per + "[" + minShuffleFrac + " - " + highShuffleFrac + "]");
			}

			final SymmetricBooleanArrayDistanceFunction usedEdges = new SymmetricBooleanArrayDistanceFunction(this.getNoOfElements());
			int lowerBound = 0;
			for (final ConflictVal conflictVal : conflictVals)
			{
				final int[] conflict = conflictVal.getValue();
				if (!usedEdges.get(conflict[0], conflict[1]) && !usedEdges.get(conflict[0], conflict[2]) && !usedEdges.get(conflict[1], conflict[2]))
				{
					lowerBound++;
					usedEdges.set(conflict[0], conflict[1], true);
					usedEdges.set(conflict[0], conflict[2], true);
					usedEdges.set(conflict[1], conflict[2], true);
				}
			}
			// if (lowerBound > lowerBoundGlobal)
			// System.err.println("new lowerBound:" + lowerBound + " (" + rounds + " " + (rounds % modShuffle == 0) + ")");

			lowerBoundGlobal = Math.max(lowerBoundGlobal, lowerBound);

			if (watch.getTimeInSec() > maxTime)
			{
				// System.err.println("break with #rounds:" + rounds + "(" + watch.getTimeInSec() + "s)");
				break;
			}
		}
		return lowerBoundGlobal;
	}

	@Override
	public boolean solve() throws UnsolvableException
	{
		throw new NotImplementedException();
	}

}
